import itertools

_registration_id = itertools.count()


class Registration(object):

    def __init__(self, func, hook):
        super(Registration, self).__init__()
        self.id = next(_registration_id)
        self.hook = hook
        self.func = func
        if not hasattr(self.func, "gossip"):
            self.func.gossip = self

    def unregister(self):
        if self.hook is not None:
            self.hook.unregister(self)
            assert self.hook is None

    def can_be_called(self):
        return True

    def __call__(self, *args, **kwargs):
        return self.func(*args, **kwargs)



import itertools

import gossip
import pytest

hook_id = itertools.count()

@pytest.fixture(autouse=True, scope="function")
def clear_registrations():
    gossip.unregister_all()
    gossip.undefine_all()

@pytest.fixture
def hook_name():
    return "hook{0}".format(next(hook_id))

@pytest.fixture
def registered_hook(hook_name):
    return RegisteredHook(hook_name)

@pytest.fixture
def registered_hooks():
    returned = []
    for i in range(10):
        if i % 3 == 0:
            hook_name = "group{0}.hook{1}".format(next(hook_id), next(hook_id))
        else:
            hook_name = "hook{0}".format(next(hook_id))
        returned.append(RegisteredHook(hook_name))
    return returned

class RegisteredHook(object):

    def __init__(self, hook_name):
        super(RegisteredHook, self).__init__()

        self.name = hook_name
        self.kwargs = {"a": 1, "b": 2, "c": 3}
        self.num_called = 0

        def handler(**kw):
            assert kw == self.kwargs
            self.num_called += 1

        self.func = handler

        gossip.register(func=handler, hook_name=hook_name)

    def works(self):
        old_num_caled = self.num_called
        gossip.trigger(self.name, **self.kwargs)
        return self.num_called == old_num_caled + 1



from .hook import Hook


class Group(object):

    def __init__(self, name):
        super(Group, self).__init__()
        self.name = name
        self._hooks = {}
        self._subgroups = {}

    def create_hook(self, name):
        assert name not in self._hooks
        hook = Hook(self, name)
        self._hooks[name] = hook
        return hook

    def get_or_create_subgroup(self, name):
        returned = self._subgroups.get(name)
        if returned is None:
            returned = self._subgroups[name] = Group(name)
        return returned

    def remove_all_children(self):
        self._hooks.clear()

    def __repr__(self):
        return "<Gossip group {0!r}>".format(self.name)



import gossip
import pytest
from gossip.exceptions import NameAlreadyUsed


def test_register_trigger(registered_hook):
    assert registered_hook.works()

def test_register_function_already_has_gossip_attr(hook_name):
    def func():
        pass

    func.gossip = old_obj = object()

    gossip.register(func=func, hook_name=hook_name)
    assert func.gossip is old_obj

def test_unregister(registered_hook):
    assert registered_hook.works()
    registered_hook.func.gossip.unregister()
    assert not registered_hook.works()

def test_unregister_all(registered_hooks):
    assert all(r.works() for r in registered_hooks)
    gossip.unregister_all()
    assert all(not r.works() for r in registered_hooks)

def test_unregister_all_specific(registered_hooks):
    assert all(r.works() for r in registered_hooks)
    gossip.unregister_all(registered_hooks[0].name)
    assert not registered_hooks[0].works()
    assert all(r.works() for r in registered_hooks[1:])

def test_unregister_all_does_not_deletes_group(registered_hooks):
    assert gossip.get_groups()
    gossip.unregister_all()
    assert gossip.get_groups()

def test_undefine_all_deletes_groups(registered_hooks):
    assert gossip.get_groups()
    gossip.undefine_all()
    assert not gossip.get_groups()
    assert all(not r.works() for r in registered_hooks)

def test_global_group_is_same(registered_hooks):
    global_group = gossip.get_global_group()
    gossip.unregister_all()
    assert gossip.get_global_group() is global_group

def test_hook_name_taken_by_group():
    @gossip.register("group_name.hook")
    def func():
        pass

    with pytest.raises(NameAlreadyUsed):
        @gossip.register("group_name")
        def func2():
            pass

    assert "group_name" not in gossip.registry._hooks

def test_group_name_taken_by_hook():
    @gossip.register("group_name")
    def func():
        pass

    with pytest.raises(NameAlreadyUsed):
        @gossip.register("group_name.hook")
        def func2():
            pass

    assert "group_name" not in gossip.registry._groups



import logging
import sys

from ._compat import reraise
from .registration import Registration

_logger = logging.getLogger(__name__)


class Hook(object):

    def __init__(self, group, name, arg_names=(), doc=None):
        super(Hook, self).__init__()
        self.group = group
        self.name = name
        self._registrations = []
        self._arg_names = arg_names
        self._swallow_exceptions = False
        self.doc = doc

    def get_argument_names(self):
        return self._arg_names

    def __call__(self, **kwargs):
        return self.trigger(kwargs)

    def register(self, func):
        self._registrations.append(Registration(func, self))

    def unregister(self, registration):
        assert registration.hook is self
        self._registrations.remove(registration)
        registration.hook = None

    def unregister_all(self):
        del self._registrations[:]

    def trigger(self, kwargs):
        last_exc_info = None
        for callback in self._registrations:
            exc_info = self._call_callback(callback, kwargs)
            if last_exc_info is None:
                last_exc_info = exc_info

        if last_exc_info and not self._swallow_exceptions:
            _logger.debug("Reraising first exception in callback")
            reraise(*last_exc_info)  # pylint: disable=W0142

    def _call_callback(self, callback, kwargs):
        exc_info = None
        try:
            callback(**kwargs)  # pylint: disable=star-args
        except:
            exc_info = sys.exc_info()
            _logger.warn("Exception occurred while calling %s",
                         callback, exc_info=exc_info)
            # TODO: debug here
        return exc_info

    def __repr__(self):
        return "<Hook {0}({1})>".format(self.name, ", ".join(self._arg_names))



import functools

from ._compat import iteritems, itervalues, string_types
from .exceptions import NameAlreadyUsed
from .group import Group

_hooks = {}

_groups = {
    None: Group("**global**")
}

def register(func, hook_name=None):
    if isinstance(func, string_types):
        return functools.partial(register, hook_name=func)
    assert hook_name is not None
    get_or_create_hook(hook_name).register(func)
    return func

def unregister_all(hook_name=None):
    hook_names = [hook_name] if hook_name is not None else _hooks
    for name in hook_names:
        _hooks[name].unregister_all()

def undefine_all():
    _hooks.clear()
    global_group = get_global_group()
    _groups.clear()
    _groups[None] = global_group
    global_group.remove_all_children()

def trigger(hook_name, **kwargs):
    hook = _hooks.get(hook_name)
    if hook is not None:
        hook.trigger(kwargs)


def get_or_create_hook(hook_name):
    returned = _hooks.get(hook_name)
    if returned is None:
        returned = _hooks[hook_name] = create_hook(hook_name)
    return returned

def create_hook(hook_name):
    if hook_name in _groups:
        raise NameAlreadyUsed("A group named {0} already exists. Cannot create a hook with the same name".format(hook_name))
    parent = get_global_group()
    group_name = None
    parts = hook_name.split(".")
    for part in parts[:-1]:
        group_name = part if group_name is None else "{0}.{1}".format(group_name, part)
        group = _groups.get(group_name)
        if group is None:
            if group_name in _hooks:
                raise NameAlreadyUsed("A hook named {0} already exists. Cannot create group with the same name".format(group_name))
            group = _groups[group_name] = parent.get_or_create_subgroup(part)
        parent = group

    hook = _hooks[hook_name] = parent.create_hook(parts[-1])
    return hook

def get_global_group():
    return _groups[None]

def get_groups():
    return list(group for group_name, group in iteritems(_groups) if group_name is not None)

def get_group_by_name(name):
    return _groups[name]
